import re
from abc import ABCMeta, abstractmethod
from itertools import islice
from operator import attrgetter
from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    from wsgi_lineprof.stats import LineProfilerStat


class BaseFilter(metaclass=ABCMeta):
    @abstractmethod
    def filter(
        self, stats: Iterable["LineProfilerStat"]
    ) -> Iterable["LineProfilerStat"]:
        pass


class FilenameFilter(BaseFilter):
    """Filter which matches with filename"""

    def __init__(self, filename: str, regex: bool = False) -> None:
        self.filename = filename
        self.regex = regex
        self.compiled_regex = re.compile(filename)

    def filter(
        self, stats: Iterable["LineProfilerStat"]
    ) -> Iterable["LineProfilerStat"]:
        if self.regex:
            compiled_regex = self.compiled_regex
            return filter(lambda s: compiled_regex.search(s.filename), stats)
        else:
            return filter(lambda s: self.filename in s.filename, stats)


class NameFilter(BaseFilter):
    """Filter which matches with name"""

    def __init__(self, name: str, regex: bool = True) -> None:
        self.name = name
        self.regex = regex
        self.compiled_regex = re.compile(name)

    def filter(
        self, stats: Iterable["LineProfilerStat"]
    ) -> Iterable["LineProfilerStat"]:
        if self.regex:
            compiled_regex = self.compiled_regex
            return filter(lambda s: compiled_regex.search(s.name), stats)
        else:
            return filter(lambda s: self.name in s.name, stats)


class TotalTimeSorter(BaseFilter):
    """Sort stats by total time"""

    def __init__(self, reverse: bool = True) -> None:
        self.reverse = reverse

    def filter(
        self, stats: Iterable["LineProfilerStat"]
    ) -> Iterable["LineProfilerStat"]:
        return sorted(stats, key=attrgetter("total_time"), reverse=self.reverse)


class TopItemsFilter(BaseFilter):
    """Get first n stats"""

    def __init__(self, n: int = 10) -> None:
        self.n = n

    def filter(
        self, stats: Iterable["LineProfilerStat"]
    ) -> Iterable["LineProfilerStat"]:
        return islice(stats, self.n)



"""Benchmarks with different filters

base: Application without LineProfilerMiddleware
no_filter: Application with LineProfilerMiddleware using no filters
filter: Application with LineProfilerMiddleware using filters
"""

from webtest import TestApp

from wsgi_lineprof.filters import FilenameFilter
from wsgi_lineprof.middleware import LineProfilerMiddleware

from .apps import fib_app, jinja_app
from .utils import StringNoopIO


class BaseTest:
    param_names = ["filters"]
    params = ["base", "no_filter", "filter"]

    def prepare_app(self, app, profiler, filters):
        if profiler == "no_filter":
            app = LineProfilerMiddleware(app, stream=StringNoopIO(), async_stream=True)
        elif profiler == "filter":
            app = LineProfilerMiddleware(
                app, filters=filters, stream=StringNoopIO(), async_stream=True
            )
        return TestApp(app)


class FibAppWithFilenameFilterTest(BaseTest):
    def setup(self, filters):
        self.app = self.prepare_app(
            fib_app,
            filters,
            filters=[
                FilenameFilter("apps.py"),
            ],
        )

    def time_index_page(self, *args):
        self.app.get("/")

    def time_n_10(self, *args):
        self.app.get("/10")

    def time_n_20(self, *args):
        self.app.get("/20")


class JinjaAppWithFilenameFilterTest(BaseTest):
    def setup(self, filters):
        self.app = self.prepare_app(
            jinja_app,
            filters,
            filters=[
                FilenameFilter("apps.py"),
            ],
        )

    def time_10_items(self, *args):
        self.app.get("/10")

    def time_100_items(self, *args):
        self.app.get("/20")



import time

from wsgi_lineprof.profiler import LineProfiler


class TestLineProfiler:
    def test_get_timer(self):
        timer = LineProfiler.get_timer()
        assert timer >= 0

    def test_get_timer_is_correct(self):
        unit = LineProfiler.get_unit()
        time1 = time.time()
        timer1 = LineProfiler.get_timer()
        time.sleep(1)
        timer2 = LineProfiler.get_timer()
        time2 = time.time()
        assert abs((timer2 - timer1) * unit - (time2 - time1)) < 0.1

    def test_get_timer_is_monotonic(self):
        timer1 = LineProfiler.get_timer()
        timer2 = LineProfiler.get_timer()
        assert timer1 <= timer2

    def test_get_timer_implementation(self):
        implementation = LineProfiler.get_timer_implementation()
        assert isinstance(implementation, str)
        assert len(implementation) > 0

    def test_get_unit(self):
        unit = LineProfiler.get_unit()
        assert isinstance(unit, float)
        assert unit > 0



from cpython cimport PyObject


cdef extern from "Python.h":
    ctypedef struct PyCodeObject:
        int       co_argcount
        int       co_nlocals
        int       co_stacksize
        int       co_flags
        PyObject *co_code
        PyObject *co_consts
        PyObject *co_names
        PyObject *co_varnames
        PyObject *co_freevars
        PyObject *co_cellvars
        PyObject *co_filename
        PyObject *co_name
        int       co_firstlineno
        PyObject *co_lnotab

    ctypedef struct PyFrameObject:
        PyFrameObject *f_back
        PyCodeObject  *f_code
        PyObject *f_builtins
        PyObject *f_globals
        PyObject *f_locals
        PyObject *f_trace
        PyObject *f_exc_type
        PyObject *f_exc_value
        PyObject *f_exc_traceback
        int f_lasti
        int f_lineno
        int f_restricted
        int f_iblock
        int f_nlocals
        int f_ncells
        int f_nfreevars
        int f_stacksize

cdef extern from "frameobject.h":
    # Once we drop Python 3.8 support, we should be able to move this definition under Python.h
    ctypedef int (*Py_tracefunc)(object self, PyFrameObject *py_frame, int what, PyObject *arg)

cdef extern from "Python.h":
    cdef void PyEval_SetTrace(Py_tracefunc func, object arg)

    # New in version 3.9
    # Part of the Stable ABI since version 3.10
    # https://docs.python.org/3/c-api/frame.html#c.PyFrame_GetCode
    cdef PyCodeObject *PyFrame_GetCode(PyFrameObject *frame)

    # Part of the Stable ABI since version 3.10
    # https://docs.python.org/3/c-api/frame.html#c.PyFrame_GetLineNumber
    cdef int PyFrame_GetLineNumber(PyFrameObject *frame)

    cdef int PyTrace_LINE
    cdef int PyTrace_RETURN



"""Benchmarks with different stream writers

base: Application without LineProfilerMiddleware
sync: Application with LineProfilerMiddleware using SyncWriter
async: Application with LineProfilerMiddleware using AsyncWriter
"""

import random
import string

from webtest import TestApp

from wsgi_lineprof.middleware import LineProfilerMiddleware

from .apps import demo_app, fib_app, jinja_app, re_app
from .utils import StringNoopIO


class BaseTest:
    param_names = ["writer"]
    params = ["base", "sync", "async"]

    def prepare_app(self, app, writer):
        if writer == "sync":
            app = LineProfilerMiddleware(app, stream=StringNoopIO())
        elif writer == "async":
            app = LineProfilerMiddleware(app, stream=StringNoopIO(), async_stream=True)
        return TestApp(app)


class DemoAppTest(BaseTest):
    def setup(self, writer):
        self.app = self.prepare_app(demo_app, writer)

    def time_index_page(self, *args):
        self.app.get("/")


class FibAppTest(BaseTest):
    def setup(self, writer):
        self.app = self.prepare_app(fib_app, writer)

    def time_index_page(self, *args):
        self.app.get("/")

    def time_n_10(self, *args):
        self.app.get("/10")

    def time_n_20(self, *args):
        self.app.get("/20")


class JinjaAppTest(BaseTest):
    def setup(self, writer):
        self.app = self.prepare_app(jinja_app, writer)

    def time_10_items(self, *args):
        self.app.get("/10")

    def time_100_items(self, *args):
        self.app.get("/20")


class ReAppTest(BaseTest):
    def setup(self, writer):
        self.app = self.prepare_app(re_app, writer)
        chars = string.ascii_uppercase + string.ascii_lowercase + string.digits
        random.seed(12345)
        random_str = "".join(random.choice(chars) for _ in range(2000))
        self.random_str = "/" + random_str
        random.seed()

    def time_index_page(self, *args):
        self.app.get("/")

    def time_email_address(self, *args):
        self.app.get("/EMAIL-address-8324236rq@example.com")

    def time_ip_address(self, *args):
        self.app.get("/ip-address-is-123.21.13.149")

    def time_random(self, *args):
        self.app.get(self.random_str)



from wsgiref.simple_server import make_server
from wsgiref.util import setup_testing_defaults

from wsgi_lineprof.filters import FilenameFilter, TotalTimeSorter
from wsgi_lineprof.middleware import LineProfilerMiddleware


def fib(n):
    if n <= 1:
        return n
    return fib(n - 1) + fib(n - 2)


# Simple WSGI application
def app(environ, start_response):
    setup_testing_defaults(environ)

    status = "200 OK"
    headers = [("Content-type", "text/plain; charset=utf-8")]

    start_response(status, headers)

    n = 30
    fib_n = fib(n)
    res = "fib(%d) = %d" % (n, fib_n)
    return [res.encode("utf-8")]


if __name__ == "__main__":
    # Set up profiler
    filters = [
        FilenameFilter("fib.py"),
        TotalTimeSorter(),
    ]
    app = LineProfilerMiddleware(app, filters=filters)

    server = make_server("127.0.0.1", 8000, app)
    print("Serving on 127.0.0.1:8000...")
    server.serve_forever()



import bottle

from wsgi_lineprof.middleware import LineProfilerMiddleware


@bottle.route("/hello/<name>")
def index(name):
    return bottle.template("<b>Hello {{name}}</b>!", name=name)


app = LineProfilerMiddleware(bottle.app())

if __name__ == "__main__":
    bottle.run(host="localhost", port=8080, app=app)



"""Benchmarks with different line profilers

base: Application without LineProfiler
noop: Application with NoopLineProfiler
enabled: Application with LineProfiler
"""

import random
import string

from webtest import TestApp

from wsgi_lineprof.extensions import LineProfiler

from .apps import demo_app, fib_app, jinja_app, re_app


class NoopLineProfiler(LineProfiler):
    def enable(self):
        self._enable_noop()


class Middleware:
    def __init__(self, app, profiler_class):
        self.app = app
        self.profiler_class = profiler_class

    def __call__(self, env, start_response):
        profiler = self.profiler_class()
        profiler.enable()
        try:
            response = self.app(env, start_response)
        finally:
            profiler.disable()
        return response


class BaseTest:
    param_names = ["profiler"]
    params = ["base", "noop", "enabled"]

    def prepare_app(self, app, profiler):
        if profiler == "noop":
            app = Middleware(app, profiler_class=NoopLineProfiler)
        elif profiler == "enabled":
            app = Middleware(app, profiler_class=LineProfiler)
        return TestApp(app)


class DemoAppTest(BaseTest):
    def setup(self, profiler):
        self.app = self.prepare_app(demo_app, profiler)

    def time_index_page(self, *args):
        self.app.get("/")


class FibAppTest(BaseTest):
    def setup(self, profiler):
        self.app = self.prepare_app(fib_app, profiler)

    def time_index_page(self, *args):
        self.app.get("/")

    def time_n_10(self, *args):
        self.app.get("/10")

    def time_n_20(self, *args):
        self.app.get("/20")


class JinjaAppTest(BaseTest):
    def setup(self, profiler):
        self.app = self.prepare_app(jinja_app, profiler)

    def time_10_items(self, *args):
        self.app.get("/10")

    def time_100_items(self, *args):
        self.app.get("/20")


class ReAppTest(BaseTest):
    def setup(self, profiler):
        self.app = self.prepare_app(re_app, profiler)
        chars = string.ascii_uppercase + string.ascii_lowercase + string.digits
        random.seed(12345)
        random_str = "".join(random.choice(chars) for _ in range(2000))
        self.random_str = "/" + random_str
        random.seed()

    def time_index_page(self, *args):
        self.app.get("/")

    def time_email_address(self, *args):
        self.app.get("/EMAIL-address-8324236rq@example.com")

    def time_ip_address(self, *args):
        self.app.get("/ip-address-is-123.21.13.149")

    def time_random(self, *args):
        self.app.get(self.random_str)



from copy import copy

import pytest

from wsgi_lineprof.extensions import LineTiming


def func():
    return 123


def func2():
    return "hello"


class TestLineTiming:
    def test_as_tuple(self):
        subject = LineTiming(func.__code__, 10)
        assert subject.as_tuple() == (10, 0, 0)

    def test_copy(self):
        timing = LineTiming(func.__code__, 1)
        timing.total_time = 2
        timing.n_hits = 3
        copied = copy(timing)
        copied.total_time = 4
        copied.n_hits = 5
        assert timing.total_time == 2
        assert timing.n_hits == 3

    def test_add_succeeds(self):
        a = LineTiming(func.__code__, 1)
        a.total_time = 2
        a.n_hits = 3
        b = LineTiming(func.__code__, 1)
        b.total_time = 4
        b.n_hits = 5
        c = a + b
        assert a.code == func.__code__
        assert a.lineno == 1
        assert a.total_time == 2
        assert a.n_hits == 3
        assert b.code == func.__code__
        assert b.lineno == 1
        assert b.total_time == 4
        assert b.n_hits == 5
        assert c.code == func.__code__
        assert c.lineno == 1
        assert c.total_time == 6
        assert c.n_hits == 8

    def test_add_not_implemented(self):
        a = LineTiming(func.__code__, 1)
        with pytest.raises(TypeError):
            a + 1

    def test_add_value_error(self):
        a = LineTiming(func.__code__, 1)
        b = LineTiming(func2.__code__, 1)
        c = LineTiming(func.__code__, 2)
        with pytest.raises(ValueError):
            a + b
        with pytest.raises(ValueError):
            a + c



import uuid
from datetime import datetime
from types import CodeType
from typing import Dict, TextIO

from typing_extensions import TypedDict

from wsgi_lineprof.extensions import LineTiming

CodeTiming = Dict[int, LineTiming]
Measurement = Dict[CodeType, CodeTiming]
RequestMeasurement = TypedDict(
    "RequestMeasurement",
    {
        "id": uuid.UUID,
        "started_at": datetime,
        "elapsed": float,
        "unit": float,
        "results": Measurement,
        "request_method": str,
        "path_info": str,
        "query_string": str,
    },
)

# TODO: Improve definition
Stream = TextIO



from types import CodeType
from typing import Callable, Iterable, Union

from wsgi_lineprof.filters import BaseFilter
from wsgi_lineprof.types import CodeTiming, Measurement, RequestMeasurement


class LineProfilerStat:
    def __init__(
        self,
        code: CodeType,
        timings: CodeTiming,
    ) -> None:
        self.code = code
        self.timings = timings
        self.total_time = sum(t.total_time for t in timings.values())

    @property
    def filename(self) -> str:
        # TODO: Improve how to handle empty filename
        filename = self.code.co_filename
        return filename if filename is not None else ""

    @property
    def name(self) -> str:
        return self.code.co_name

    @property
    def firstlineno(self) -> int:
        return self.code.co_firstlineno


CallableFilterType = Callable[[Iterable[LineProfilerStat]], Iterable[LineProfilerStat]]
FilterType = Union[CallableFilterType, BaseFilter]


class LineProfilerStats:
    def __init__(self, stats: Iterable[LineProfilerStat], unit: float) -> None:
        self.stats = stats
        self.unit = unit  # seconds/hit

    @classmethod
    def from_request_measurement(
        cls, request_measurement: RequestMeasurement
    ) -> "LineProfilerStats":
        return cls.from_measurement_and_unit(
            request_measurement["results"], request_measurement["unit"]
        )

    @classmethod
    def from_measurement_and_unit(
        cls, measurement: Measurement, unit: float
    ) -> "LineProfilerStats":
        return cls([LineProfilerStat(c, t) for c, t in measurement.items()], unit)

    def filter(self, f: FilterType) -> "LineProfilerStats":
        if isinstance(f, BaseFilter):
            return LineProfilerStats(f.filter(self.stats), self.unit)
        else:
            return LineProfilerStats(f(self.stats), self.unit)



from collections import OrderedDict
from datetime import datetime
from uuid import UUID

import pytest
from pytz import utc

from wsgi_lineprof.app import ResultsApp

SHOULD_HANDLE_REQUEST_TEST_CASES = [
    ("/wsgi_lineprof/", "/wsgi_lineprof/", True),
    ("/wsgi_lineprof/", "/wsgi_lineprof/hello", True),
    ("/wsgi_lineprof/", "", False),
    ("/wsgi_lineprof/", "/hello", False),
    ("/wsgi_lineprof/", "/wsgi_lineprof", False),
]

REQUEST_1 = {
    "id": UUID("1967c955-838f-41f4-b78f-6b4adfccbaf2"),
    "started_at": datetime(2020, 1, 5, 17, 12, 53, 123, tzinfo=utc),
    "elapsed": 0.321,
    "unit": 0.001,
    "results": {},
    "request_method": "GET",
    "path_info": "/fib",
    "query_string": "",
}


class TestResultsApp:
    @pytest.fixture
    def results(self):
        return OrderedDict()

    @pytest.fixture
    def start_response(self, mocker):
        return mocker.Mock()

    @pytest.fixture
    def app(self, results):
        return ResultsApp(endpoint="/wsgi_lineprof/", results=results, filters=[])

    def test_init_enforces_trailing_slash(self, results):
        app = ResultsApp(endpoint="/wsgi_lineprof", results=results, filters=[])
        assert app.endpoint == "/wsgi_lineprof/"

    @pytest.mark.parametrize(
        "endpoint,path_info,expected", SHOULD_HANDLE_REQUEST_TEST_CASES
    )
    def test_should_handle_request(self, app, endpoint, path_info, expected):
        app.endpoint = endpoint
        env = {"PATH_INFO": path_info}
        assert app.should_handle_request(env) == expected

    def test_routing(self, mocker, app, start_response):
        env = {"PATH_INFO": "/wsgi_lineprof/"}
        handler = mocker.spy(app, "_handle_index")
        app(env, start_response)
        handler.assert_called_once_with(start_response)

    def test_index_with_empty_results(self, app, start_response, results):
        env = {"PATH_INFO": "/wsgi_lineprof/"}

        response = b"".join(app(env, start_response)).decode("utf-8")

        start_response.assert_called_once_with(
            "200 OK", [("Content-Type", "text/html; charset=utf-8")]
        )
        assert len(response) > 0
        assert "GET" not in response

    def test_index_with_results(self, app, start_response, results):
        env = {"PATH_INFO": "/wsgi_lineprof/"}
        results[REQUEST_1["id"]] = REQUEST_1

        response = b"".join(app(env, start_response)).decode("utf-8")

        start_response.assert_called_once_with(
            "200 OK", [("Content-Type", "text/html; charset=utf-8")]
        )
        assert len(response) > 0
        assert "GET" in response
        assert "/fib" in response
        assert "321" in response

    def test_detail_with_results(self, app, start_response, results):
        env = {"PATH_INFO": "/wsgi_lineprof/1967c955-838f-41f4-b78f-6b4adfccbaf2"}
        results[REQUEST_1["id"]] = REQUEST_1

        response = b"".join(app(env, start_response)).decode("utf-8")

        start_response.assert_called_once_with(
            "200 OK", [("Content-Type", "text/html; charset=utf-8")]
        )
        assert len(response) > 0
        assert "GET" in response
        assert "/fib" in response
        assert "321" in response

    def test_detail_without_results(self, app, start_response, results):
        env = {"PATH_INFO": "/wsgi_lineprof/1967c955-838f-41f4-b78f-0123456789ab"}

        response = b"".join(app(env, start_response)).decode("utf-8")

        start_response.assert_called_once_with("404 Not Found", [])
        assert len(response) == 0

    def test_detail_with_invalid_url(self, app, start_response, results):
        env = {"PATH_INFO": "/wsgi_lineprof/abcdefg"}

        response = b"".join(app(env, start_response)).decode("utf-8")

        start_response.assert_called_once_with("404 Not Found", [])
        assert len(response) == 0

    def test_not_found(self, app, start_response, results):
        env = {"PATH_INFO": "/hello"}

        response = b"".join(app(env, start_response)).decode("utf-8")

        start_response.assert_called_once_with("404 Not Found", [])
        assert len(response) == 0



from sys import version_info

from Cython.Build import cythonize
from setuptools import Extension, setup

setup(
    ext_package="wsgi_lineprof",
    ext_modules=cythonize(
        [
            Extension(
                "extensions",
                sources=["extensions/extensions.pyx", "extensions/timer.c"],
            )
        ],
        compile_time_env={
            "PY_MAJOR_VERSION": version_info.major,
            "PY_MINOR_VERSION": version_info.minor,
        },
    ),
)



import re
from io import StringIO
from typing import TYPE_CHECKING, Any, Dict, Iterable
from uuid import UUID

from jinja2 import Environment, PackageLoader

from wsgi_lineprof.formatter import TextFormatter
from wsgi_lineprof.stats import FilterType, LineProfilerStats
from wsgi_lineprof.types import RequestMeasurement
from wsgi_lineprof.writers import SyncStreamWriter

if TYPE_CHECKING:
    from wsgiref.types import StartResponse, WSGIEnvironment


UUID_RE = re.compile(
    "^[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}$", re.I
)


class ResultsApp:
    """WSGI application to show the profiling results"""

    def __init__(
        self,
        endpoint: str,
        results: Dict[UUID, RequestMeasurement],
        filters: Iterable[FilterType],
    ) -> None:
        assert endpoint.startswith("/")
        if not endpoint.endswith("/"):
            endpoint += "/"
        self.endpoint = endpoint
        self.results = results
        self.filters = filters
        self.template_env = Environment(
            loader=PackageLoader("wsgi_lineprof", "templates"), autoescape=True
        )

    def __call__(
        self, env: "WSGIEnvironment", start_response: "StartResponse"
    ) -> Iterable[bytes]:
        if not self.should_handle_request(env):
            return _return_404(start_response)

        path = env["PATH_INFO"][len(self.endpoint) :]
        if path == "":
            return self._handle_index(start_response)
        elif UUID_RE.match(path):
            return self._handle_detail(start_response, UUID(path))
        else:
            return _return_404(start_response)

    def should_handle_request(self, env: "WSGIEnvironment") -> bool:
        """Returns true only if this app should handle the request"""
        return bool(env["PATH_INFO"].startswith(self.endpoint))

    def _handle_index(self, start_response: "StartResponse") -> Iterable[bytes]:
        template = self.template_env.get_template("index.html")
        # To suppress the following mypy error on Python 3:
        # error: No overload variant of "reversed" matches argument
        results: Any = self.results.values()
        start_response("200 OK", [("Content-Type", "text/html; charset=utf-8")])
        return [template.render(results=reversed(results)).encode("utf-8")]

    def _handle_detail(
        self, start_response: "StartResponse", request_id: UUID
    ) -> Iterable[bytes]:
        request_measurement = self.results.get(request_id)
        if request_measurement is None:
            return _return_404(start_response)

        template = self.template_env.get_template("detail.html")
        stream: Any = StringIO()
        writer = SyncStreamWriter(stream, TextFormatter(color=False))
        stats = LineProfilerStats.from_request_measurement(request_measurement)
        for f in self.filters:
            stats = stats.filter(f)
        writer.write(stats)
        start_response("200 OK", [("Content-Type", "text/html; charset=utf-8")])
        return [
            template.render(result=request_measurement, stats=stream.getvalue()).encode(
                "utf-8"
            )
        ]


def _return_404(start_response: "StartResponse") -> Iterable[bytes]:
    start_response("404 Not Found", [])
    return []



from abc import ABCMeta, abstractmethod
from queue import Queue
from threading import Thread
from typing import Any, Optional

from wsgi_lineprof.formatter import BaseFormatter
from wsgi_lineprof.stats import LineProfilerStats
from wsgi_lineprof.types import Stream


class BaseStreamWriter(metaclass=ABCMeta):
    def __init__(
        self,
        stream: Stream,
        formatter: BaseFormatter,
        *kwargs: Any,
    ) -> None:
        self.stream = stream
        self.formatter = formatter

    @abstractmethod
    def write(self, stats: LineProfilerStats) -> None:
        return


class SyncStreamWriter(BaseStreamWriter):
    def write(self, stats: LineProfilerStats) -> None:
        self.formatter.format_stats(stats, self.stream)


class AsyncStreamWriter(BaseStreamWriter):
    def __init__(
        self,
        stream: Stream,
        formatter: BaseFormatter,
    ) -> None:
        super(AsyncStreamWriter, self).__init__(stream, formatter)
        # None in the queue is a special message to stop the writer thread
        self.queue: Queue[Optional[LineProfilerStats]] = Queue()
        self.writer_thread = Thread(target=self._write)
        self.writer_thread.daemon = True
        self.writer_thread.start()

    def write(self, stats: LineProfilerStats) -> None:
        # Avoid accidentally stopping the writer thread by writing None
        if stats is None:
            return
        self.queue.put(stats)

    def _write(self) -> None:
        """Method to run in the writer thread"""
        while True:
            stats = self.queue.get()
            if stats is None:
                return
            self.formatter.format_stats(stats, self.stream)

    def _join(self) -> None:
        """Utility method to stop the writer thread

        Mainly for testing.
        """
        self.queue.put(None)
        self.writer_thread.join()



from wsgiref.simple_server import demo_app, make_server

from wsgi_lineprof.middleware import LineProfilerMiddleware

app = LineProfilerMiddleware(demo_app)

if __name__ == "__main__":
    with make_server("", 8000, app) as httpd:
        print("Serving HTTP on port 8000...")
        httpd.serve_forever()



import threading

from wsgi_lineprof.stats import LineProfilerStats
from wsgi_lineprof.writers import AsyncStreamWriter, SyncStreamWriter


class TestSyncStreamWriter:
    def test_write_calls_format_stats(self, mocker):
        stream = mocker.Mock()
        formatter = mocker.Mock()
        writer = SyncStreamWriter(stream, formatter)
        stats = LineProfilerStats([], 0.001)

        writer.write(stats)

        formatter.format_stats.assert_called_once_with(stats, stream)


class TestAsyncStreamWriter:
    def test_write_calls_format_stats(self, mocker):
        stream = mocker.Mock()
        formatter = mocker.Mock()
        writer = AsyncStreamWriter(stream, formatter)
        stats = LineProfilerStats([], 0.001)

        writer.write(stats)
        writer._join()

        formatter.format_stats.assert_called_once_with(stats, stream)

    def test_write_when_none_is_given(self, mocker):
        stream = mocker.Mock()
        formatter = mocker.Mock()
        writer = AsyncStreamWriter(stream, formatter)

        writer.write(None)
        writer._join()

        formatter.format_stats.assert_not_called()

    def test_join_stops_thread(self, mocker):
        original_active_thread_count = threading.active_count()
        writer = AsyncStreamWriter(mocker.Mock(), mocker.Mock())
        assert original_active_thread_count + 1 == threading.active_count()

        writer._join()

        assert original_active_thread_count == threading.active_count()
